local addonName, addonTable = ...

local craftsDB = {}
addonTable.craftsDB = craftsDB

local myRealm = GetRealmName()
local myFaction = UnitFactionGroup("player")
local myName = UnitName("player")
local updatedBags = {}
local recipeColorCache = {}
local menu = {}
local addonLoaded = {}
local options, newCraftMsgShown, bankOpen, lastRecipe, colorRecipesFunc

local ERR_SKILL_GAINED_S	= ERR_SKILL_GAINED_S:format("(.+)")		-- "You have gained the (.+) skill."
local ERR_LEARN_RECIPE_S	= ERR_LEARN_RECIPE_S:format("(.+)%")	-- "You have learned how to create a new item: (.+)%."
local ERR_SPELL_UNLEARNED_S	= ERR_SPELL_UNLEARNED_S:format("(.+)%")	-- "You have unlearned (.+)%."
local ERR_SKILL_UP_SI		= format("^%s$", ERR_SKILL_UP_SI:gsub("%%[sd]", "(%0+)"):gsub("%%s", "."))			-- "^Your skill in (.+) has increased to (%d+).$"
local ITEM_MIN_SKILL		= format("^%s$", ITEM_MIN_SKILL:gsub("%%s", ".+"):gsub("%(%%d%)", "%%((%%d+)%%)"))	-- "^Requires .+ %((.+)%)$"
local ITEM_REQ_SKILL		= format("^%s$", ITEM_REQ_SKILL:format("(.+)"))										-- "^Requires (.+)$"

-- used for checking item type and subtype returns from GetItemInfo
local AH_RECIPE_INDEX = 7
local RECIPE = select(AH_RECIPE_INDEX, GetAuctionItemClasses())
local BOOK = GetAuctionItemSubClasses(AH_RECIPE_INDEX)

-- "apprentice" rank spell IDs, used for localization
local skills = {
	2018, -- Blacksmithing
	2108, -- Leatherworking
	2259, -- Alchemy
	2550, -- Cooking
	-- 2575, -- Mining
	3273, -- First Aid
	3908, -- Tailoring
	4036, -- Engineering
	7411, -- Enchanting
	25229, -- Jewelcrafting
	45357, -- Inscription
}

-- this lets us lookup trade skill spell IDs by name, eg:
-- skillsByName["Blacksmithing"] = 2018
-- used for getting trade skill ID from recipes and the trade skill frame
local skillsByName = {}

for i = 1, #skills do
	local spellID = skills[i]
	local spellName = GetSpellInfo(spellID)
	skillsByName[spellName] = spellID
	skills[spellID] = spellName
	skills[i] = nil
end

local specs = {
	[20219] = 4036, -- Gnomish Engineer
	[20222] = 4036, -- Goblin Engineer
	-- [28677] = 2259, -- Elixir Master
}

local specsByName = {}

for spellID in pairs(specs) do
	specsByName[GetSpellInfo(spellID)] = spellID
end


local L = {
	["Usable"]				= "Usable",
	["Insufficient skill"]	= "Insufficient skill",
	["Already known"]		= "Already known",
	["Unuseful recipe"]		= "Unuseful recipe",

	["In bags: "]	= "In bags: ",
	["In bank: "]	= "In bank: ",
	
	BC_MSG_DATA_DELETED	= ": data deleted",
	["Please open your tradeskill/craft to update data"]	= "Please open your tradeskill/craft to update data",
	
	["Delete character data"] = "Delete character data",
	["Color recipes"] = "Color recipes",
	["Bags"] = "Bags",
	["Bank"] = "Bank",
	["Guild bank"] = "Guild bank",
	["Vendors"] = "Vendors",
	["Auction house"] = "Auction house",

	["Only same faction"]			= "Only same faction",
	["Show skill level"]			= "Show skill level",
	["Include current character"]	= "Include current character",
	["Use side tooltip"]			= "Use side tooltip",
	
	["Are you sure you want to reset data for %s @ %s?"] = "Are you sure you want to reset data for %s @ %s?",
	["Are you sure you want to reset %s data for %s?"] = "Are you sure you want to reset %s data for %s?",
}

local locale = GetLocale()

if locale == "frFR" then
	L["Usable"]				= "Peut apprendre"
	L["Insufficient skill"]	= "Rang insuffisant"
	L["Already known"]		= "A dj appris"
	L["Unuseful recipe"]	= "Recette inutile"

	-- L["In bags: "]	= "In bags: "
	L["In bank: "]	= "A la banque: "
	
	BC_MSG_DATA_DELETED	= ": donnes effaces"
	L["Please open your tradeskill/craft to update data"]	= "Merci de bien vouloir ouvrir la fentre d'artisanat pour une mise  jour."

	-- L["Delete character data"] = "Delete character data"
	-- L["Color recipes"] = "Color recipes"
	-- L["Bags"] = "Bags"
	-- L["Bank"] = "Bank"
	-- L["Guild bank"] = "Guild bank"
	-- L["Vendors"] = "Vendors"
	-- L["Auction house"] = "Auction house"

	L["Only same faction"]			= "Personnages de la mme faction"
	L["Show skill level"]			= "Niveau de comptence"
	L["Include current character"]	= "Personnage courant"
	L["Use side tooltip"]			= "Tooltip indpendant"
	
	-- L["Are you sure you want to reset data for %s @ %s?"] = "Are you sure you want to reset data for %s @ %s?"
	-- L["Are you sure you want to reset %s data for %s?"] = "Are you sure you want to reset %s data for %s?"
elseif locale == "deDE" then
	L["Usable"]				= "Gleich erlernbar"
	L["Insufficient skill"]	= "Spter erlernbar"
	L["Already known"]		= "Bereits bekannt"
	L["Unuseful recipe"]	= "Nicht erlernbar / Fertigkeit noch nicht erlernt"

	-- L["In bags: "]	= "In bags: "
	-- L["In bank: "]	= "In bank: "
	
	-- BC_MSG_DATA_DELETED	= ": data deleted"
	-- L["Please open your tradeskill/craft to update data"]	= "Please open your tradeskill/craft to update data"

	-- L["Delete character data"] = "Delete character data"
	-- L["Color recipes"] = "Color recipes"
	-- L["Bags"] = "Bags"
	-- L["Bank"] = "Bank"
	-- L["Guild bank"] = "Guild bank"
	-- L["Vendors"] = "Vendors"
	-- L["Auction house"] = "Auction house"

	L["Only same faction"]			= "Nur Charakter der eigenen Fraktion anzeigen"
	L["Show skill level"]			= "Fertigkeitsrnge anzeigen"
	L["Include current character"]	= "Auch diesen Charakter bei der Auswertung miteinbeziehen"
	L["Use side tooltip"]			= "Seitlichen Tooltip benutzen"
	
	-- L["Are you sure you want to reset data for %s @ %s?"] = "Are you sure you want to reset data for %s @ %s?"
	-- L["Are you sure you want to reset %s data for %s?"] = "Are you sure you want to reset %s data for %s?"
end

local addColon = {
	"Usable",
	"Insufficient skill",
	"Already known",
}

for i, v in ipairs(addColon) do
	L[v..": "] = L[v]..": "
end


local addon = CreateFrame("Frame")
addon:RegisterEvent("ADDON_LOADED")
addon:RegisterEvent("PLAYER_LOGIN")
addon:RegisterEvent("UNIT_NAME_UPDATE")
addon:RegisterEvent("TRADE_SKILL_UPDATE")
addon:RegisterEvent("TRADE_SKILL_SHOW")
addon:RegisterEvent("LEARNED_SPELL_IN_TAB")
addon:RegisterEvent("BANKFRAME_OPENED")
addon:RegisterEvent("BANKFRAME_CLOSED")
addon:RegisterEvent("CHAT_MSG_SKILL")
addon:RegisterEvent("CHAT_MSG_SYSTEM")
addon:SetScript("OnEvent", function(self, event, ...)
	self[event](self, ...)
end)

addon.name = "Book of Crafts"
InterfaceOptions_AddCategory(addon)

local title = addon:CreateFontString(nil, nil, "GameFontNormalLarge")
title:SetPoint("TOPLEFT", 16, -16)
title:SetPoint("RIGHT", -16, 0)
title:SetJustifyH("LEFT")
title:SetJustifyV("TOP")
title:SetText("Book of Crafts")


local function onClick(self)
	SetItemRef(self.value:match("|H(.-)|h"), self.value)
end

local dropdown = CreateFrame("Frame", "BoCLDBMenu")
dropdown.displayMode = "MENU"
function dropdown:initialize(level)
	if level == 1 then
		for i, v in ipairs(menu[myRealm].charNames) do
			if v ~= myName then
				local info = UIDropDownMenu_CreateInfo()
				info.text = v
				info.hasArrow = true
				info.notCheckable = true
				UIDropDownMenu_AddButton(info, level)
			end
		end
	elseif level == 2 then
		for k, v in pairs(addon.realm[UIDROPDOWNMENU_MENU_VALUE].skills) do
			local info = UIDropDownMenu_CreateInfo()
			info.text = skills[k]
			info.value = v.link
			info.func = onClick
			info.notCheckable = true
			UIDropDownMenu_AddButton(info, level)
		end
	end
end

local dataobj = LibStub("LibDataBroker-1.1"):NewDataObject(addonName, {
	type = "launcher",
	icon = [[Interface\Icons\INV_Misc_Book_09]],
	label = "Book of Crafts",
	OnClick = function(self)
		ToggleDropDownMenu(nil, nil, dropdown, self, 0, 0)
	end,
})


local function onClick(self)
	if self:GetChecked() then
		PlaySound("igMainMenuOptionCheckBoxOn")
	else
		PlaySound("igMainMenuOptionCheckBoxOff")
	end
	options[self.setting] = self:GetChecked()
end

local function newOption(parent)
	local button = CreateFrame("CheckButton", nil, parent, "OptionsBaseCheckButtonTemplate")
	button:SetPushedTextOffset(0, 0)
	button:SetScript("OnClick", onClick)
	
	local text = button:CreateFontString(nil, nil, "GameFontHighlight")
	text:SetPoint("LEFT", button, "RIGHT", 0, 1)
	button:SetFontString(text)
	
	return button
end


local function setColor(id)
	local r, g, b = ColorPickerFrame:GetColorRGB()
	local swatch = addon.swatches[id]
	swatch.tx:SetVertexColor(r, g, b)
	local color = swatch.color
	color.r = r
	color.g = g
	color.b = b
end

local function cancel(id, prev)
	local r, g, b = prev.r, prev.g, prev.b
	local swatch = addon.swatches[id]
	swatch.tx:SetVertexColor(r, g, b)
	local color = swatch.color
	color.r = r
	color.g = g
	color.b = b
end

local function onClick(self)
	local info = UIDropDownMenu_CreateInfo()
	local color = self.color
	info.r, info.g, info.b = color.r, color.g, color.b
	info.swatchFunc = self.swatchFunc
	info.cancelFunc = self.cancelFunc
	OpenColorPicker(info)
end

local function onEnter(self)
	self.bg:SetVertexColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
end

local function onLeave(self)
	self.bg:SetVertexColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
end

local function newSwatch(parent)
	local swatch = CreateFrame("Button", nil, parent)
	swatch:SetSize(16, 16)
	swatch:SetPushedTextOffset(0, 0)
	
	swatch:SetNormalTexture([[Interface\ChatFrame\ChatFrameColorSwatch]])
	swatch.tx = swatch:GetNormalTexture()
	
	local bg = swatch:CreateTexture(nil, "BACKGROUND")
	bg:SetTexture(1.0, 1.0, 1.0)
	bg:SetSize(14, 14)
	bg:SetPoint("CENTER")
	swatch.bg = bg
	
	local text = swatch:CreateFontString(nil, nil, "GameFontHighlight")
	text:SetPoint("LEFT", swatch, "RIGHT", 5, 0)
	text:SetJustifyH("LEFT")
	swatch:SetFontString(text)
	
	swatch:SetScript("OnClick", onClick)
	swatch:SetScript("OnEnter", onEnter)
	swatch:SetScript("OnLeave", onLeave)
	
	return swatch
end


do
	local options = {
		{
			text = L["Include current character"],
			setting = "showPlayer",
		},
		{
			text = L["Only same faction"],
			setting = "sameFaction",
		},
		{
			text = L["Show skill level"],
			setting = "showSkillLevel",
		},
		{
			text = L["Use side tooltip"],
			setting = "separateTooltip",
		},
	}
	
	for i, v in ipairs(options) do
		local option = newOption(addon)
		if i == 1 then
			option:SetPoint("TOPLEFT", title, "BOTTOMLEFT", -2, -16)
		else
			option:SetPoint("TOP", options[i - 1], "BOTTOM", 0, -8)
		end
		option:SetText(v.text)
		option.setting = v.setting
		options[i] = option
	end
	
	addon.options = options
end


do
	local swatches = {
		{
			text = L["Usable"],
			setting = "usable",
		},
		{
			text = L["Insufficient skill"],
			setting = "lowSkill",
		},
		{
			text = L["Already known"],
			setting = "knownBy",
		},
		{
			text = L["Unuseful recipe"],
			setting = "useless",
		},
	}
	
	for i, v in ipairs(swatches) do
		local swatch = newSwatch(addon)
		swatch:SetID(i)
		if i == 1 then
			swatch:SetPoint("TOP", addon.options[#addon.options], "BOTTOM", 0, -13)
		else
			swatch:SetPoint("TOP", swatches[i - 1], "BOTTOM", 0, -18)
		end
		swatch:SetText(v.text)
		swatch.setting = v.setting
		swatch.swatchFunc = function() setColor(i) end
		swatch.cancelFunc = function(prev) cancel(i, prev) end
		swatches[i] = swatch
	end
	
	addon.swatches = swatches
end


local colorRecipesMenu = {
	"bags",
	"bank",
	"guildBank",
	"vendor",
	"auction",
}

local colorRecipesMenuLabels = {
	bags = "Bags",
	bank = "Bank",
	guildBank = "Guild bank",
	vendor = "Vendors",
	auction = "Auction house",
}

local function onClick(self)
	options.colorRecipes[self.value] = self.checked
	colorRecipesFunc[self.value]()
end

local colorRecipes = CreateFrame("Frame", "BoCColorRecipesDropdown", addon, "UIDropDownMenuTemplate")
colorRecipes:SetPoint("TOPLEFT", addon.swatches[#addon.swatches], "BOTTOMLEFT", -16, -18)
UIDropDownMenu_SetWidth(colorRecipes, 160)
UIDropDownMenu_SetText(colorRecipes, L["Color recipes"])
function colorRecipes:initialize()
	for i, v in ipairs(colorRecipesMenu) do
		local info = UIDropDownMenu_CreateInfo()
		info.text = colorRecipesMenuLabels[v]
		info.value = v
		info.func = onClick
		info.checked = options.colorRecipes[v]
		info.isNotRadio = true
		info.keepShownOnClick = true
		UIDropDownMenu_AddButton(info)
	end
end


local popupData = {}

local function onClick(self, realm, name)
	local skill = self.value
	popupData.realm = realm
	popupData.name = name
	popupData.skill = skill
	UIDropDownMenu_SetSelectedValue(self.owner, self.value)
	skill = skills[skill]
	if skill then
		StaticPopup_Show("BOOKOFCRAFTS_RESET_SKILL", skills[skill], name.." @ "..realm, popupData)
	else
		StaticPopup_Show("BOOKOFCRAFTS_RESET_CHARACTER", name, realm, popupData)
	end
end

local charDropdown = CreateFrame("Frame", "BoCUI_DropDownCharacters", addon, "UIDropDownMenuTemplate")
charDropdown:SetPoint("TOP", colorRecipes, "BOTTOM", 0, -8)
UIDropDownMenu_SetWidth(charDropdown, 160)
UIDropDownMenu_SetText(charDropdown, L["Delete character data"])

charDropdown.initialize = function(self, level)
	if level == 1 then
		for i, v in ipairs(menu.realmNames) do
			local info = UIDropDownMenu_CreateInfo()
			info.text = v
			info.hasArrow = true
			info.notCheckable = true
			UIDropDownMenu_AddButton(info, level)
		end
	elseif level == 2 then
		local realm = UIDROPDOWNMENU_MENU_VALUE
		for i, v in ipairs(menu[realm].charNames) do
			local info = UIDropDownMenu_CreateInfo()
			info.text = v
			info.value = realm.."|"..v
			info.func = onClick
			info.hasArrow = true
			info.notCheckable = true
			info.owner = self
			info.arg1 = realm
			info.arg2 = v
			UIDropDownMenu_AddButton(info, level)
		end
	elseif level == 3 then
		local realm, name = strsplit("|", UIDROPDOWNMENU_MENU_VALUE)
		for i, v in ipairs(menu[realm].data[name]) do
			local info = UIDropDownMenu_CreateInfo()
			info.text = skills[v]
			info.value = v
			info.func = onClick
			info.notCheckable = true
			info.arg1 = realm
			info.arg2 = name
			UIDropDownMenu_AddButton(info, level)
		end
	end
end

StaticPopupDialogs["BOOKOFCRAFTS_RESET_CHARACTER"] = {
	text = L["Are you sure you want to reset data for %s @ %s?"],
	button1 = OKAY,
	button2 = CANCEL,
	OnAccept = function(self, data)
		addon:DeleteCharacterData(data.realm, data.name)
		UIDropDownMenu_SetText(charDropdown, "Delete character data")
	end,
	OnCancel = function(self, data)
		UIDropDownMenu_SetText(charDropdown, "Delete character data")
	end,
	whileDead = true,
	timeout = 0,
}

StaticPopupDialogs["BOOKOFCRAFTS_RESET_SKILL"] = {
	text = L["Are you sure you want to reset %s data for %s?"],
	button1 = OKAY,
	button2 = CANCEL,
	OnAccept = function(self, data)
		addon:DeleteCharacterData(data.realm, data.name)
	end,
	whileDead = true,
	timeout = 0,
}


local tooltip = CreateFrame("GameTooltip", "BoCTooltip", UIParent, "ShoppingTooltipTemplate")
-- _G[self:GetName().."TextRight1"]:SetFontObject("GameFontNormalSmall")
BoCTooltipTextLeft2:SetFontObject("GameFontNormalSmall")
tooltip:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b)
tooltip:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b)
local addLine = tooltip.AddLine
function tooltip:AddLine(text, color)
	if color then
		addLine(self, text, color.r, color.g, color.b)
	else
		addLine(self, text, 1.0, 1.0, 1.0)
	end
end


-------------------------------------------------------------------------------
-- Feedback functions
-------------------------------------------------------------------------------

function addon:ChatMsg(msg)
	DEFAULT_CHAT_FRAME:AddMessage("[Book of Crafts] "..msg, 0.0, 1.0, 0.0)
end


SLASH_BookOfCrafts1 = "/boc"
SLASH_BookOfCrafts2 = "/bookofcrafts"
SlashCmdList.BookOfCrafts = function(cmd)
	InterfaceOptionsFrame_OpenToCategory(addon)
end


local function onUpdate(self, elapsed)
	self:ScanInventory()
	self:SetScript("OnUpdate", nil)
end


--[[

DB = {
	["Char|Realm"] = {
		skills = {
			spellID = {
				crafts = {
					-- any learnt trade skill crafts
					[craftID] = true,
					[...]
				},
				-- any learnt specialisation of the trade skill such as gnomish engineering, etc
				spec = spellID,
				link = "link",
				skillLevel = tradeSkillLevel,
			},
		},
		-- recipes this character possess
		recipes = {
			bags = {
				[itemID] = amount,
			},
			bank = {
				[itemID] = amount,
			},
		},
		faction = "Faction",
	},
}

]]

local defaults = {
	options = {
		sameFaction = true,
		showSkillLevel = true,
		showPlayer = true,
		debug = false,
		colorRecipes = {
			bags = true,
			bank = true,
			vendor = true,
			auction = true,
			guildBank = true,
		},
		colors = {
			knownBy		= {r = 0.5, g = 0.5, b = 1.0},
			lowSkill	= {r = 1.0, g = 0.0, b = 0.0},
			usable	= {r = 0.0, g = 1.0, b = 0.0},
			useless		= {r = 0.5, g = 0.5, b = 0.5},
		},
	},
	data = {},
}

local function copyDefaults(src, dst)
	if not src then return {} end
	if not dst then dst = {} end
	for k, v in pairs(src) do
		if type(v) == "table" then
			dst[k] = copyDefaults(v, dst[k])
		elseif type(v) ~= type(dst[k]) then
			dst[k] = v
		end
	end
	return dst
end

function addon:ADDON_LOADED(addon)
	if self:HookAddon(addon) then
		return
	end
	
	if addon == addonName then
		local db = BookofCraftsDB or {}
		copyDefaults(defaults, db)
		
		BookofCraftsDB = db
		options = db.options

		self.realm = db.data[myRealm] or {}
		db.data[myRealm] = self.realm
		
		self.player = self.realm[myName] or {
			skills = {},
			recipes = {
				bags = {},
				bank = {},
			},
			faction = myFaction,
		}
		self.realm[myName] = self.player
		
		--[[
		
		menu = {
			realmNames = {
				-- alphabetically sorted array
				"Azjol-Nerub",
				"Bronzebeard",
				"Chromaggus",
			}
			["Azjol-Nerub"] = {
				charNames = {
					"Armina",
					"Belleza",
					"Cordiana",
				},
				data = {
					["Armina"] = {
						skills = {
							2108,
							3273,
						},
					},
				},
			},
			["Bronzebeard"] = {
			},
			["Chromaggus"] = {
			},
		}
		
		]]
		
		-- sorted-by-name array for the LDB dropdown
		local realmNames = {}
		for realmName, realmData in pairs(db.data) do
			tinsert(realmNames, realmName)
			local charNames = {}
			local chars = {}
			for charName, charData in pairs(realmData) do
				tinsert(charNames, charName)
				local skills = {}
				for spellID in pairs(charData.skills) do
					tinsert(skills, spellID)
				end
				sort(skills)
				chars[charName] = skills
			end
			sort(charNames)
			menu[realmName] = {
				charNames = charNames,
				data = chars,
			}
		end
		sort(realmNames)
		menu.realmNames = realmNames
		
		for i, v in ipairs(self.options) do
			v:SetChecked(options[v.setting])
		end
		
		for i, swatch in pairs(self.swatches) do
			local color = options.colors[swatch.setting]
			swatch.tx:SetVertexColor(color.r, color.g, color.b)
			swatch.color = color
		end
		
		GameTooltip:HookScript("OnTooltipSetItem", self.AnalyseTooltip)
		GameTooltip:HookScript("OnTooltipCleared", function(self)
			self.BoC = nil
		end)
		GameTooltip:HookScript("OnHide", function(self)
			BoCTooltip:Hide()
		end)

		self:HookAddons()
	end
end


function addon:PLAYER_LOGIN()
	self:RegisterEvent("BAG_UPDATE")
	self:ScanProfessions()
end


function addon:UNIT_NAME_UPDATE()
	self:ScanProfessions()
end


function addon:TRADE_SKILL_SHOW()
	newCraftMsgShown = false
end


function addon:TRADE_SKILL_UPDATE()
	if not (IsTradeSkillLinked() or IsTradeSkillGuild()) then
		-- return if not yet fully loaded
		if select(2, GetTradeSkillInfo(1)) ~= "header" then return end
		
		local tradeSkill, skillLevel = GetTradeSkillLine()
		tradeSkill = skillsByName[tradeSkill]

		if tradeSkill then
			tradeSkill = self.player.skills[tradeSkill]

			tradeSkill.link = GetTradeSkillListLink()
			tradeSkill.skillLevel = skillLevel
			local newCraft
			local crafts = tradeSkill.crafts

			for i = 1, GetNumTradeSkills() do
				local recipeLink = GetTradeSkillRecipeLink(i)
				
				if recipeLink then
					local craftID = tonumber(recipeLink:match("enchant:(%d+)"))
					if not crafts[craftID] then
						crafts[craftID] = true
						newCraft = true
					end
				end
			end
			if newCraft then
				self:ClearCache()
			end
		end
	end
end


function addon:LEARNED_SPELL_IN_TAB(spellID)
	if specs[spellID] then
		self:ScanProfessions()
	end
end


function addon:BAG_UPDATE()
	self:SetScript("OnUpdate", onUpdate)
end


function addon:BANKFRAME_OPENED()
	bankOpen = true
	self:ScanInventory()
end


function addon:BANKFRAME_CLOSED()
	bankOpen = false
end


function addon:CHAT_MSG_SKILL(msg)
	local tradeSkillName, newLevel = msg:match(ERR_SKILL_UP_SI)
	local tradeSkill = self.player.skills[skillsByName[tradeSkillName]]

	if tradeSkill and newLevel then
		-- newLevel = tonumber(newLevel)

		tradeSkill.skillLevel = tonumber(newLevel)
		tradeSkill.link = select(2, GetSpellLink(tradeSkillName))
		self:ClearCache()
	end
	
	local tradeskill = msg:match(ERR_SKILL_GAINED_S)
	
	if tradeskill and skillsByName[tradeskill] then
		self:ScanProfessions()
	end
end


local sysMsgPatterns = {
	[ERR_LEARN_RECIPE_S] = function()
		if not newCraftMsgShown then
			newCraftMsgShown = true
			addon:ChatMsg(L["Please open your tradeskill/craft to update data"])
		end
	end,
	[ERR_SPELL_UNLEARNED_S] = function(match)
		local spellID = tonumber(match:match("spell:(%d+)"))
		if skills[spellID] then
			-- this spell ID matches a valid tradeskill, so we'll delete it from the database
			addon.player.skills[spellID] = nil
		else
			-- this spell ID matches a valid tradeskill spec, so we'll delete the spec entry
			local skill = specs[spellID]
			if skill then
				addon.player.skills[skill].spec = nil
			end
		end
	end,
}

function addon:CHAT_MSG_SYSTEM(msg)
	for pattern, func in pairs(sysMsgPatterns) do
		local match = msg:match(pattern)
		if match then
			func(match)
			break
		end
	end
end


function addon:ScanProfessions()
	for i = 1, select("#", GetProfessions()) do
		local prof = select(i, GetProfessions())
		if prof then
			local name, _, skillLevel, _, numAbilities, spellOffset = GetProfessionInfo(prof)
			local skillID = skillsByName[name]
			if skillID then
				local skills = self.player.skills
				skills[skillID] = skills[skillID] or {crafts = {}}
				local _, link = GetSpellLink(skillID)
				-- links are not available on first PLAYER_LOGIN, in which case we wait for UNIT_NAME_UPDATE
				if not link then
					return
				end
				skills[skillID].skillLevel = skillLevel
				skills[skillID].link = link
				if numAbilities == 2 then
					-- we have two profession abilities, one of them may be a profession spec
					skills[skillID].spec = specsByName[GetSpellInfo(spellOffset + 2, BOOKTYPE_PROFESSION)]
				end
			end
		end
	end
	-- if we reached this point, the links were available and we no longer need the following event
	self:UnregisterEvent("UNIT_NAME_UPDATE")
end


local scanTooltip = CreateFrame("GameTooltip", "BoCScanTooltip", nil, "GameTooltipTemplate")

function addon:GetRecipeInfo(itemLink)
	local skillLevel, specialReq

	scanTooltip:SetOwner(UIParent, "ANCHOR_NONE")
	scanTooltip:SetHyperlink(itemLink)
	-- parse required skill level and potential profession spec
	for i = 2, 3 do
		local text = _G["BoCScanTooltipTextLeft"..i]:GetText()
		if text then
			skillLevel = text:match(ITEM_MIN_SKILL)
			if skillLevel then
				break
			end
		end
	end
	for i = 3, 4 do
		local text = _G["BoCScanTooltipTextLeft"..i]:GetText()
		if text then
			local req = text:match(ITEM_REQ_SKILL)
			if specsByName[req] then
				specialReq = req
				break
			end
		end
	end
	return tonumber(itemLink:match("item:(%d+)")), skillsByName[self:IsRecipe(itemLink)], tonumber(skillLevel), specialReq
end


function addon:IsRecipe(itemLink)
	local type, subType = select(6, GetItemInfo(itemLink))
	return (type == RECIPE and subType ~= BOOK and subType)
end


function addon:ScanBag(bag, src)
	for slot = 1, GetContainerNumSlots(bag) do
		local itemID = GetContainerItemID(bag, slot)
		local _, count, _, _, _, _, link = GetContainerItemInfo(bag, slot)
		if itemID and self:IsRecipe(link) then
			local recipes = self.player.recipes[src]
			recipes[itemID] = (recipes[itemID] or 0) + count
		end
	end
end


function addon:ScanInventory()
	-- Clear character "InBank" list
	local recipes = self.player.recipes
	recipes.bags = {}
	
	for bag = 0, 4 do
		self:ScanBag(bag, "bags")
	end

	if bankOpen then
		recipes.bank = {}
		self:ScanBag(BANK_CONTAINER, "bank")
		-- Search for recipes in bank bags
		for bag = 5, 11 do
			self:ScanBag(bag, "bank")
		end
	end
end


local cache = {
	usable = {},
	lowSkill = {},
	knownBy = {},
	bags = {},
	bank = {},
}

function addon:GetAltsInfo(itemLink)
	local usable, lowSkill, knownBy, bags, bank = cache.usable, cache.lowSkill, cache.knownBy, cache.bags, cache.bank

	-- if this recipe is the same as the last scanned recipe, we can just use the cached data
	if itemLink == lastRecipe then
		return usable, lowSkill, knownBy, bags, bank
	end

	for k, v in pairs(cache) do
		wipe(v)
	end
	local itemID, tradeSkill, skillLevel, specialReq = self:GetRecipeInfo(itemLink)
	local crafts = craftsDB[tradeSkill]

	for charName, v in pairs(self.realm) do
		local recipes = v.recipes
		local inBags = recipes.bags[itemID]
		local inBank = recipes.bank[itemID]
		bags[#bags + 1] = inBags and format("%s (%d)", charName, inBags)
		bank[#bank + 1] = inBank and format("%s (%d)", charName, inBank)
		local skillData = v.skills[tradeSkill]
		if skillData --[[and (options.showPlayer or charName ~= myName)]] and (not options.sameFaction or v.faction == myFaction) then
			if skillData.crafts[crafts[itemID]] then
				-- if this recipe exists in the character's database, then the character knows it
				knownBy[#knownBy + 1] = charName
			elseif not specialReq or specsByName[specialReq] == skillData.spec then
				-- if no spec is required, or the character has that spec, then they'll be able to learn this recipe..
				local text = charName
				if options.showSkillLevel then
					text = format("%s (%d)", text, skillData.skillLevel)
				end
				if skillData.skillLevel >= skillLevel then
					-- ..now, since the character's skill level is high enough..
					usable[#usable + 1] = text
				else
					-- ..or later
					lowSkill[#lowSkill + 1] = text
				end
			end
		end
	end

	return usable, lowSkill, knownBy, bags, bank
end


-- add extra tooltip information if necessary
function addon:AnalyseTooltip()
	local _, itemLink = self:GetItem()
	if not self.BoC and itemLink then
		-- Analyse tooltip data if we are dealing with a tradeskill/craft object
		if addon:IsRecipe(itemLink) then
			-- Local results only needs an update if it is not the same object than last one analysed
			-- if lastRecipe ~= itemLink then
				addon:ShowResults(self)
				lastRecipe = itemLink
				self.BoC = true
			-- end
		else
			-- Record that no object is analysed
			lastRecipe = nil
		end
	end
end


local isUseful

local function dostuff(table, caption, color)
	if #table > 0 then
		BoCTooltip:AddLine(caption, color)
		for i, v in ipairs(table) do
			BoCTooltip:AddLine("  "..v, color)
		end
		isUseful = true
	end
end

local function dostuff2(tooltip, table, caption, color)
	if #table > 0 then
		tooltip:AddLine(caption..table.concat(table, ", "), color.r, color.g, color.b)
		isUseful = true
	end
end

function addon:ShowResults(tooltip)
	local colors = options.colors

	local usable, lowSkill, knownBy, bags, bank = self:GetAltsInfo(select(2, tooltip:GetItem()))
	isUseful = nil
	
	if options.separateTooltip then
		BoCTooltip:SetOwner(tooltip, "ANCHOR_NONE")

		dostuff(usable, L["Usable"], colors.usable)
		dostuff(lowSkill, L["Insufficient skill"], colors.lowSkill)
		dostuff(knownBy, L["Already known"], colors.knownBy)

		if not isUseful then
			BoCTooltip:AddLine(L["Unuseful recipe"])
		end

		dostuff(bags, L["In bags: "])
		dostuff(bank, L["In bank: "])

		addon:TooltipShow(tooltip)
	else
		tooltip:AddLine(" ")

		dostuff2(tooltip, usable, L["Usable: "], colors.usable)
		dostuff2(tooltip, lowSkill, L["Insufficient skill: "], colors.lowSkill)
		dostuff2(tooltip, knownBy, L["Already known: "], colors.knownBy)

		if not isUseful then
			tooltip:AddLine(L["Unuseful recipe"], 1.0, 1.0, 1.0)
		end

		dostuff2(tooltip, bags, L["In bags: "])
		dostuff2(tooltip, bank, L["In bank: "])
	end
end


function addon:TooltipShow(frame)
	local tooltipWidth = tooltip:GetWidth()
	local left = (frame:GetLeft() or 0) - tooltipWidth
	local right = GetScreenWidth() - (frame:GetRight() or 0) - tooltipWidth
	if left < right then
		tooltip:SetPoint("TOPLEFT", frame, "TOPRIGHT")
	else
		tooltip:SetPoint("TOPRIGHT", frame, "TOPLEFT")
	end

	local color = TOOLTIP_DEFAULT_BACKGROUND_COLOR
	tooltip:SetBackdropColor(color.r, color.g, color.b)
	tooltip:Show()
end


function addon:DeleteCharacterData(realm, name, skill)
	local data = BookofCraftsDB.data[realm]
	if skill then
		local skills = data[name].skills
		wipe(skills[skill].crafts)
		skills[skill] = nil
	else
		data[name] = nil
	end
	self:ClearCache()
	self:ChatMsg(realm..name..(skill and skills[skill] or "")..BC_MSG_DATA_DELETED)
end


function addon:ClearCache()
	lastRecipe = nil
	wipe(recipeColorCache)
end


local red = {r = 1.0, g = 0, b = 0}

function addon:GetRecipeColor(itemLink)
	local color = recipeColorCache[itemLink]
	if not color then
		local usable, lowSkill, knownBy = self:GetAltsInfo(itemLink)
		color = options.colors.useless
		if #usable > 0 then
			color = options.colors.usable
		elseif #knownBy > 0 then
			color = options.colors.knownBy
		elseif #lowSkill > 0 then
			color = options.colors.lowSkill
		end
		recipeColorCache[itemLink] = color
	end
	return color
end


-- function addon:ColorRecipes(loc)
	-- if options.colorRecipes[loc] then
		-- colorRecipesFunc[loc]()
	-- end
-- end


local function colorBagRecipes(self)
	if not self.size then
		return
	end
	local id = self:GetID()
	local bagName = self:GetName()
	for i = 1, self.size do
		local button = _G[bagName.."Item"..i]
		local slot = button:GetID()
		local link = GetContainerItemLink(id, slot)
		if options.colorRecipes.bags and link and addon:IsRecipe(link) then
			local color = addon:GetRecipeColor(link)
			SetItemButtonTextureVertexColor(button, color.r, color.g, color.b)
		else
			SetItemButtonTextureVertexColor(button, 1.0, 1.0, 1.0)
		end
	end
end

local function colorBankRecipes(self)
	local link = GetInventoryItemLink("player", self:GetInventorySlot())
	if options.colorRecipes.bank and link and addon:IsRecipe(link) then
		local color = addon:GetRecipeColor(link)
		SetItemButtonTextureVertexColor(self, color.r, color.g, color.b)
	else
		SetItemButtonTextureVertexColor(self, 1.0, 1.0, 1.0)
	end
end

local function colorGuildBankRecipes()
	if GuildBankFrame.mode == "bank" then
		local tab = GetCurrentGuildBankTab()
		for i = 1, MAX_GUILDBANK_SLOTS_PER_TAB do
			local index = mod(i, NUM_SLOTS_PER_GUILDBANK_GROUP)
			if index == 0 then
				index = NUM_SLOTS_PER_GUILDBANK_GROUP
			end
			local column = ceil((i - 0.5) / NUM_SLOTS_PER_GUILDBANK_GROUP)
			local button = _G["GuildBankColumn"..column.."Button"..index]
			local link = GetGuildBankItemLink(tab, i)
			if options.colorRecipes.guildBank and link and addon:IsRecipe(link) then
				local color = addon:GetRecipeColor(link)
				SetItemButtonTextureVertexColor(button, color.r, color.g, color.b)
			else
				SetItemButtonTextureVertexColor(button, 1.0, 1.0, 1.0)
			end
		end
	end
end

local function colorVendorRecipes()
	if not options.colorRecipes.vendor then return end
	for i = 1, min(MERCHANT_ITEMS_PER_PAGE, GetMerchantNumItems()) do
		local index = ((MerchantFrame.page - 1) * MERCHANT_ITEMS_PER_PAGE) + i
		local itemButton = _G["MerchantItem"..i.."ItemButton"]
		local merchantButton = _G["MerchantItem"..i]
		local link = GetMerchantItemLink(index)
		if not link then return end
		if addon:IsRecipe(link) then
			local _, _, _, _, numAvailable, isUsable = GetMerchantItemInfo(index)
			local color = addon:GetRecipeColor(link)
			local r, g, b = color.r, color.g, color.b
			-- SetItemButtonNameFrameVertexColor(merchantButton, r, g, b)
			if numAvailable == 0 then
				r, g, b = r * 0.5, g * 0.5, b * 0.5
				SetItemButtonSlotVertexColor(merchantButton, r, g, b)
				SetItemButtonTextureVertexColor(itemButton, r, g, b)
				SetItemButtonNormalTextureVertexColor(itemButton, r, g, b)
			else
				SetItemButtonSlotVertexColor(merchantButton, r, g, b)
				SetItemButtonTextureVertexColor(itemButton, r, g, b)
				SetItemButtonNormalTextureVertexColor(itemButton, r, g, b)
			end
		end
	end
end

local function colorVendorBuybackRecipes()
	if not options.colorRecipes.vendor then return end
	for i = 1, min(MERCHANT_ITEMS_PER_PAGE, GetNumBuybackItems()) do
		local itemButton = _G["MerchantItem"..i.."ItemButton"]
		local merchantButton = _G["MerchantItem"..i]
		local link = GetBuybackItemLink(i)
		if not link then return end
		if addon:IsRecipe(link) then
			local color = addon:GetRecipeColor(link)
			local r, g, b = color.r, color.g, color.b
			SetItemButtonNameFrameVertexColor(merchantButton, r, g, b)
			SetItemButtonSlotVertexColor(merchantButton, r, g, b)
			SetItemButtonTextureVertexColor(itemButton, r, g, b)
			SetItemButtonNormalTextureVertexColor(itemButton, r, g, b)
		end
	end
end

local function colorAuctionRecipes()
	if not options.colorRecipes.auction then return end
	for i = 1, min(NUM_BROWSE_TO_DISPLAY, GetNumAuctionItems("list")) do
		local itemLink = GetAuctionItemLink("list", i + FauxScrollFrame_GetOffset(BrowseScrollFrame))
		if itemLink and addon:IsRecipe(itemLink) then
			local color = addon:GetRecipeColor(itemLink)
			_G["BrowseButton"..i.."ItemIconTexture"]:SetVertexColor(color.r, color.g, color.b)
		end
	end
end

-- these functions are called whenever the corresponding setting is toggled
colorRecipesFunc = {
	bags = function()
		for i = 1, NUM_CONTAINER_FRAMES do
			colorBagRecipes(_G["ContainerFrame"..i])
		end
	end,
	bank = function()
		for i = 1, NUM_BANKGENERIC_SLOTS do
			BankFrameItemButton_Update(_G["BankFrameItem"..i])
		end
	end,
	guildBank = function()
		if addonLoaded.Blizzard_GuildBankUI and GuildBankFrame:IsShown() then
			GuildBankFrame_Update()
		end
	end,
	vendor = function()
		if MerchantFrame:IsShown() then
			MerchantFrame_UpdateMerchantInfo()
			-- MerchantFrame_UpdateBuybackInfo()
		end
	end,
	auction = function()
		if addonLoaded.Blizzard_AuctionUI and AuctionFrame:IsShown() then
			-- AuctionFrameBrowse_Update()
			colorAuctionRecipes()
		end
	end,
}

hooksecurefunc("ContainerFrame_Update", colorBagRecipes)
hooksecurefunc("ContainerFrame_UpdateCooldown", function(container, button)
	-- local cooldown = _G[button:GetName().."Cooldown"]
	-- local start, duration, enable = GetContainerItemCooldown(container, button:GetID())
	-- CooldownFrame_SetTimer(cooldown, start, duration, enable)
	-- if duration > 0 and enable == 0 then
		-- SetItemButtonTextureVertexColor(button, 0.4, 0.4, 0.4)
	-- else
		-- SetItemButtonTextureVertexColor(button, 1, 1, 1)
	-- end
	local link = GetContainerItemLink(container, button:GetID())
	if options.colorRecipes.bags and link and addon:IsRecipe(link) then
		local color = addon:GetRecipeColor(link)
		SetItemButtonTextureVertexColor(button, color.r, color.g, color.b)
	end
end)
hooksecurefunc("BankFrameItemButton_Update", colorBankRecipes)
hooksecurefunc("MerchantFrame_UpdateMerchantInfo", colorVendorRecipes)
hooksecurefunc("MerchantFrame_UpdateBuybackInfo", colorVendorBuybackRecipes)


local addonHooks = {
	Blizzard_AuctionUI = "AuctionFrameBrowse_Update",
	Blizzard_GuildBankUI = "GuildBankFrame_Update",
}

local hookFuncs = {
	AuctionFrameBrowse_Update = colorAuctionRecipes,
	GuildBankFrame_Update = colorGuildBankRecipes,
}

function addon:HookAddon(addon)
	local hook = addonHooks[addon]
	if hook then
		hooksecurefunc(hook, hookFuncs[hook])
		addonLoaded[addon] = true
		addonHooks[addon] = nil
		if not next(addonHooks) then
			self:UnregisterEvent("ADDON_LOADED")
			self.ADDON_LOADED = nil
		end
		return true
	end
end


function addon:HookAddons()
	for addon in pairs(addonHooks) do
		if IsAddOnLoaded(addon) then
			self:HookAddon(addon)
		end
	end
end